import { IdentitySet, Value } from 'preserves';

export type Assertion = Value<Ref>;

export type Handle = number;

export type ExitReason = null | { ok: true } | { ok: false, err: Error };

export const assert = Symbol('assert');
export const retract = Symbol('retract');
export const message = Symbol('message');

export interface Entity {
    [assert]?(turn: Turn, assertion: Assertion, handle: Handle): void;
    [retract]?(turn: Turn, handle: Handle): void;
    [message]?(turn: Turn, message: Assertion): void;
}

export class Ref {
    readonly actor: Actor;
    readonly target: Entity;

    constructor(actor: Actor, target: Entity) {
        this.actor = actor;
        this.target = target;
    }

    sync(turn: Turn, syncable: Ref) {
        turn.enqueue(syncable.actor, t => t.message(syncable, true));
    }
}

export type OutboundMap = Map<Handle, Ref>;

export class Actor {
    readonly outbound: OutboundMap;
    exitReason: ExitReason = null;

    constructor(initialAssertions: OutboundMap = new Map()) {
        this.outbound = initialAssertions;
    }

    terminateWith(t: Turn, reason: Exclude<ExitReason, null>) {
        if (this.exitReason !== null) return;
        this.exitReason = reason;
        this.outbound.forEach((peer, h) => t._retract(peer, h));
    }

    execute(proc: () => void): void {
        queueMicrotask(() => {
            if (this.exitReason !== null) return;
            try {
                proc();
            } catch (err) {
                console.error(Actor, err);
                Turn.for(this, t => this.terminateWith(t, { ok: false, err }));
            }
        });
    }
}

let nextHandle = 0;

type LocalAction = (t: Turn) => void;

export class Turn {
    readonly actor: Actor;
    readonly queues: Map<Actor, LocalAction[]> = new Map();
    readonly tasks: Array<LocalAction> = [];

    static for(actor: Actor, f: LocalAction): void {
        const t = new Turn(actor);
        f(t);
        t.queues.forEach((q, a) => a.execute(() => q.forEach(f => Turn.for(a, f))));
        t.tasks.length && queueMicrotask(() => t.tasks.forEach(f => Turn.for(actor, f)));
    }

    private constructor(actor: Actor) {
        this.actor = actor;
    }

    ref(t: Entity): Ref {
        return new Ref(this.actor, t);
    }

    spawn(bootProc: LocalAction, initialAssertions = new IdentitySet<Handle>()): void {
        this.tasks.push(() => {
            const newOutbound: OutboundMap = new Map();
            initialAssertions.forEach(key => {
                newOutbound.set(key, this.actor.outbound.get(key)!); // we trust initialAssertions
                this.actor.outbound.delete(key);
            });
            const child = new Actor(newOutbound);
            child.execute(() => Turn.for(child, bootProc));
        });
    }

    quit(): void {
        this.tasks.push(t => this.actor.terminateWith(t, { ok: true }));
    }

    assert(location: Ref, assertion: Assertion): Handle {
        const h = nextHandle++;
        this.enqueue(location.actor, t => {
            this.actor.outbound.set(h, location);
            location.target[assert]?.(t, assertion, h);
        });
        return h;
    }

    retract(h: Handle): void {
        this._retract(this.actor.outbound.get(h)!, h);
    }

    replace(location: Ref, h: Handle | undefined, assertion: Assertion): Handle {
        const newHandle = this.assert(location, assertion);
        if (h !== void 0) this.retract(h);
        return newHandle;
    }

    _retract(location: Ref, handle: Handle): void {
        this.enqueue(location.actor, t => {
            this.actor.outbound.delete(handle);
            location.target[retract]?.(t, handle);
        });
    }

    sync(loc: Ref): Promise<Turn> {
        return new Promise(resolve => this.enqueue(loc.actor, t =>
            loc.sync(t, this.ref({ [message]: resolve }))));
    }

    message(location: Ref, assertion: Assertion): void {
        this.enqueue(location.actor, t => location.target[message]?.(t, assertion));
    }

    enqueue(actor: Actor, a: LocalAction): void {
        this.queues.get(actor)?.push(a) ?? this.queues.set(actor, [a]);
    }
}
